/*
 * Copyright notice from original mutt:
 * Copyright (C) 2004 Thomas Glanzmann <sithglan@stud.uni-erlangen.de>
 * Copyright (C) 2004 Tobias Werth <sitowert@stud.uni-erlangen.de>
 * Copyright (C) 2004 Brian Fundakowski Feldman <green@FreeBSD.org>
 *
 * This file is part of mutt-ng, see http://www.muttng.org/.
 * It's licensed under the GNU General Public License,
 * please see the file GPL in the top level source directory.
 */

#include <lib-lib/lib-lib.h>

#ifdef USE_HCACHE

#if defined(HAVE_QDBM)
#include <depot.h>
#include <cabin.h>
#include <villa.h>
#elif defined(HAVE_GDBM)
#include <gdbm.h>
#else
#error neither HAVE_QDBM nor HAVE_GDBM are set ?!
#endif

#include "charset.h"
#include "mutt.h"
#include "hcache.h"

struct hcache_t {
#if defined(HAVE_QDBM)
    VILLA *db;
#elif defined(HAVE_GDBM)
    GDBM_FILE db;
#endif
    char *folder;
    unsigned int crc;
};

static unsigned int crc32(unsigned int crc, const void *src, ssize_t len)
{
    int i;
    const unsigned char *p = src;

    while (len--) {
        crc ^= *p++;
        for (i = 0; i < 8; i++)
            crc = (crc >> 1) ^ ((crc & 1) ? 0xedb88320 : 0);
    }
    return crc;
}

static int generate_crc32(void)
{
    static int crc = 0;

    crc = crc32(crc, "madmutt.2007.05.13", m_strlen("madmutt.2007.05.13"));
#ifdef HAVE_LANGINFO_CODESET
    crc = crc32(crc, MCharset.charset, m_strlen(MCharset.charset));
    crc = crc32(crc, "HAVE_LANGINFO_CODESET",
                m_strlen("HAVE_LANGINFO_CODESET"));
#endif
    crc = crc32(crc, "USE_POP",   m_strlen("USE_POP"));
    crc = crc32(crc, "MIXMASTER", m_strlen("MIXMASTER"));
    crc = crc32(crc, "USE_IMAP",  m_strlen("USE_IMAP"));
#ifdef USE_NNTP
    crc = crc32(crc, "USE_NNTP",  m_strlen("USE_NNTP"));
#endif
    return crc;
}

static const char *
mutt_hcache_per_folder(const char *path, const char *folder)
{
    static char buf[_POSIX_PATH_MAX];
    struct stat st;
    int pos;

    if (stat(path, &st) < 0 || !S_ISDIR(st.st_mode)) {
        return path;
    }

    pos  = m_strcpy(buf, sizeof(buf), path);
    pos += m_strputc(buf + pos, sizeof(buf) - pos, '/');

    while (*folder) {
        if (isalnum((unsigned char)*folder) || *folder == '.') {
            pos += m_strputc(buf + pos, sizeof(buf) - pos, *folder);
        } else {
            pos += m_strputc(buf + pos, sizeof(buf) - pos, '_');
        }
        folder++;
    }
    pos += m_strcpy(buf + pos, sizeof(buf) - pos, ".hdb");
    return buf;
}

/* store and restore things {{{ */

static const void *restore_int(const char *d, int *i)
{
    memcpy(i, d, sizeof(*i));
    return d + sizeof(*i);
}

static void dump_cstr(buffer_t *buf, const char *s)
{
    int size = 0;

    if (m_strisempty(s)) {
        buffer_add(buf, &size, sizeof(size));
        return;
    }

    size = strlen(s) + 1;
    buffer_add(buf, &size, sizeof(size));
    buffer_add(buf, s, size);
}

static const void *restore_cstr(const char *d, char **c)
{
    int size;

    d  = restore_int(d, &size);
    *c = p_dup(d, size);
    return d + size;
}

static void dump_address(buffer_t *buf, address_t *a)
{
    int counter = 0, pos = buf->len;

    buffer_add(buf, &counter, sizeof(counter));

    for (; a; a = a->next, counter++) {
        dump_cstr(buf, a->personal);
        dump_cstr(buf, a->mailbox);
        buffer_add(buf, &a->group, sizeof(a->group));
    }

    memcpy(buf->data + pos, &counter, sizeof(counter));
}

static const void *restore_address(const char *d, address_t **a)
{
    int counter;

    d = restore_int(d, &counter);

    for (; counter > 0; counter--) {
        *a = address_new();
        d = restore_cstr(d, &(*a)->personal);
        d = restore_cstr(d, &(*a)->mailbox);
        d = restore_int(d, &(*a)->group);
        a = &(*a)->next;
    }

    *a = NULL;
    return d;
}

static void dump_list(buffer_t *buf, string_list_t *l)
{
    int pos = buf->len;
    int counter = 0;

    buffer_add(buf, &counter, sizeof(counter));

    for (; l; l = l->next, counter++) {
        dump_cstr(buf, l->data);
    }

    memcpy(buf->data + pos, &counter, sizeof(counter));
}

static const void *restore_list(const char *d, string_list_t **l)
{
    int counter;

    d = restore_int(d, &counter);

    for (; counter > 0; counter--) {
        *l = string_item_new();
        d = restore_cstr(d, &(*l)->data);
        l = &(*l)->next;
    }

    *l = NULL;
    return d;
}

static void dump_parameter(buffer_t *buf, parameter_t *p)
{
    int pos = buf->len, counter = 0;

    buffer_add(buf, &counter, sizeof(counter));

    for (; p; p = p->next, counter++) {
        dump_cstr(buf, p->attribute);
        dump_cstr(buf, p->value);
    }

    memcpy(buf->data + pos, &counter, sizeof(counter));
}

static const void *restore_parameter(const char *d, parameter_t ** p)
{
    int counter;

    d = restore_int(d, &counter);

    for (; counter > 0; counter--) {
        *p = parameter_new();
        d  = restore_cstr(d, &(*p)->attribute);
        d  = restore_cstr(d, &(*p)->value);
        p  = &(*p)->next;
    }

    *p = NULL;
    return d;
}

static void dump_body(buffer_t *buf, BODY *b)
{
    buffer_add(buf, b, sizeof(*b));
    dump_cstr(buf, b->xtype);
    dump_cstr(buf, b->subtype);

    dump_parameter(buf, b->parameter);

    dump_cstr(buf, b->description);
    dump_cstr(buf, b->form_name);
    dump_cstr(buf, b->filename);
    dump_cstr(buf, b->d_filename);
}

static const void *restore_body(const char *d, BODY *c)
{
    memcpy(c, d, sizeof(*c));
    d += sizeof(*c);

    d = restore_cstr(d, &c->xtype);
    d = restore_cstr(d, &c->subtype);

    d = restore_parameter(d, &c->parameter);

    d = restore_cstr(d, &c->description);
    d = restore_cstr(d, &c->form_name);
    d = restore_cstr(d, &c->filename);
    d = restore_cstr(d, &c->d_filename);
    return d;
}

static void dump_envelope(buffer_t *buf, ENVELOPE * e)
{
    int n;

    dump_address(buf, e->return_path);
    dump_address(buf, e->from);
    dump_address(buf, e->to);
    dump_address(buf, e->cc);
    dump_address(buf, e->bcc);
    dump_address(buf, e->sender);
    dump_address(buf, e->reply_to);
    dump_address(buf, e->mail_followup_to);

    dump_cstr(buf, e->subject);
    n = e->real_subj ? e->real_subj - e->subject : -1;
    buffer_add(buf, &n, sizeof(n));

    dump_cstr(buf, e->message_id);
    dump_cstr(buf, e->supersedes);
    dump_cstr(buf, e->date);
    dump_cstr(buf, e->x_label);
    dump_cstr(buf, e->list_post);

#ifdef USE_NNTP
    dump_cstr(buf, e->newsgroups);
    dump_cstr(buf, e->xref);
    dump_cstr(buf, e->followup_to);
    dump_cstr(buf, e->x_comment_to);
#endif

    dump_list(buf, e->references);
    dump_list(buf, e->in_reply_to);
    dump_list(buf, e->userhdrs);
}

static const void *restore_envelope(const char *d, ENVELOPE *e)
{
    int real_subj_off;

    d = restore_address(d, &e->return_path);
    d = restore_address(d, &e->from);
    d = restore_address(d, &e->to);
    d = restore_address(d, &e->cc);
    d = restore_address(d, &e->bcc);
    d = restore_address(d, &e->sender);
    d = restore_address(d, &e->reply_to);
    d = restore_address(d, &e->mail_followup_to);

    d = restore_cstr(d, &e->subject);
    d = restore_int(d, &real_subj_off);
    if (real_subj_off >= 0) {
        e->real_subj = e->subject + real_subj_off;
    } else {
        e->real_subj = NULL;
    }
    d = restore_cstr(d, &e->message_id);
    d = restore_cstr(d, &e->supersedes);
    d = restore_cstr(d, &e->date);
    d = restore_cstr(d, &e->x_label);
    d = restore_cstr(d, &e->list_post);

#ifdef USE_NNTP
    d = restore_cstr(d, &e->newsgroups);
    d = restore_cstr(d, &e->xref);
    d = restore_cstr(d, &e->followup_to);
    d = restore_cstr(d, &e->x_comment_to);
#endif

    d = restore_list(d, &e->references);
    d = restore_list(d, &e->in_reply_to);
    d = restore_list(d, &e->userhdrs);

    return d;
}

static buffer_t *mutt_hcache_dump(hcache_t *db, HEADER *h, long uid_validity)
{
    buffer_t *res = buffer_new();

    if (!uid_validity) {
        uid_validity = time(NULL);
    }
    buffer_add(res, &uid_validity, sizeof(uid_validity));
    buffer_add(res, &db->crc, sizeof(db->crc));
    buffer_add(res, h, sizeof(*h));

    dump_envelope(res, h->env);
    dump_body(res, h->content);
    dump_cstr(res, h->maildir_flags);
    return res;
}

HEADER *mutt_hcache_restore(const void *_d, HEADER **oh)
{
    const char *d = _d;
    HEADER *h = header_new();

    /* skip uid_validity + crc */
    d += sizeof(long) + sizeof(int);

    memcpy(h, d, sizeof(*h));
    d += sizeof(*h);

    h->env = envelope_new();
    d = restore_envelope(d, h->env);

    h->content = body_new();
    d = restore_body(d, h->content);

    d = restore_cstr(d, &h->maildir_flags);

    /* this is needed for maildir style mailboxes */
    if (oh) {
        h->old = (*oh)->old;
        h->path = m_strdup((*oh)->path);
        header_delete(oh);
    }

    return h;
}

/* }}} */

hcache_t *mutt_hcache_open(const char *path, const char *folder)
{
    hcache_t *h = p_new(hcache_t, 1);

    h->folder = m_strdup(folder);
    h->crc = generate_crc32();

    if (m_strisempty(path)) {
        p_delete(&h->folder);
        p_delete(&h);
        return NULL;
    }

    path = mutt_hcache_per_folder(path, folder);

    {
#if defined(HAVE_QDBM)
        int flags = VL_OWRITER | VL_OCREAT;
        if (option(OPTHCACHECOMPRESS))
            flags |= VL_OZCOMP;

        h->db = vlopen(path, flags, VL_CMPLEX);
#elif defined(HAVE_GDBM)
        int pagesize = atoi(HeaderCachePageSize) ?: 16384;

        h->db = gdbm_open((char *) path, pagesize, GDBM_WRCREAT, 00600, NULL);
#endif
    }

    if (!h->db) {
        p_delete(&h->folder);
        p_delete(&h);
    }
    return h;
}

void mutt_hcache_close(hcache_t **db)
{
    if (!*db)
        return;

#if defined(HAVE_QDBM)
    vlclose((*db)->db);
#elif defined(HAVE_GDBM)
    gdbm_close((*db)->db);
#endif

    p_delete(&(*db)->folder);
    p_delete(db);
}

void *mutt_hcache_fetch(hcache_t *db, const char *filename,
                        ssize_t (*keylen)(const char *fn))
{
    char path[_POSIX_PATH_MAX];
    void *data = NULL;

    if (!db)
        return NULL;

    snprintf(path, sizeof(path), "%s%s", db->folder, filename);

    {
#if defined(HAVE_QDBM)
        int ksize = strlen(db->folder) + keylen(path + strlen(db->folder));
        data  = vlget(db->db, path, ksize, NULL);
#elif defined(HAVE_GDBM)
        datum k = { .dptr = path, .dsize = keylen(path) };

        data = gdbm_fetch(db->db, k).dtpr;
#endif
    }

    if (data) {
        unsigned crc = 0;

        restore_int(data + sizeof(long), (int *)&crc);
        if (crc != db->crc)
            p_delete(&data);
    }

    return data;
}

int mutt_hcache_store(hcache_t *db, const char *filename, HEADER *header,
                      long uid_validity, ssize_t (*keylen)(const char *fn))
{
    char path[_POSIX_PATH_MAX];
    buffer_t *data;
    int ret;

    if (!db)
        return -1;

    snprintf(path, sizeof(path), "%s%s", db->folder, filename);
    data = mutt_hcache_dump(db, header, uid_validity);

    {
#if defined(HAVE_QDBM)
        int ksize = strlen(db->folder) + keylen(path + strlen(db->folder));

        ret = vlput(db->db, path, ksize, data->data, data->len, VL_DOVER);
#elif defined(HAVE_GDBM)
        datum k = { .dptr = path, .dsize = keylen(path) };
        datum v = { .dptr = data->data, .dsize = data->len };

        ret = gdbm_store(db->db, k, v, GDBM_REPLACE);
#endif
    }

    buffer_delete(&data);
    return ret;
}

int mutt_hcache_delete(hcache_t *db, const char *filename,
                       ssize_t (*keylen)(const char *fn))
{
    char path[_POSIX_PATH_MAX];

    if (!db)
        return -1;

    snprintf(path, sizeof(path), "%s%s", db->folder, filename);

    {
#if defined(HAVE_QDBM)
        int ksize = strlen(db->folder) + keylen(path + strlen(db->folder));
        return vlout(db->db, path, ksize);
#elif defined(HAVE_GDBM)
        datum k = { .dptr = path, .dsize = keylen(path) };
        return gdbm_delete(db->db, k);
#endif
    }
}

#endif /* USE_HCACHE */
